//#if defined(__linux__)
//#define _GNU_SOURCE
//#endif // defined(__linux__)

#include <stdio.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <px_url.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <assert.h>
#include <pwd.h>
#include <arpa/inet.h>
#include <fcntl.h>

// the directory in $HOME that will be searched for known certificates
#define DEFAULT_KNOWN_CERTS_DIR ".pxc/known_certs"

// we set tofu as the highest priority here because gemini certs are often self
// signed and TOFU is the chosen way that the official gemini protocol spec
// indiates authentication should be done
enum pxc_cert_verify_mode {
  e_pxc_cert_verify_tofu = 0,   // trust on first use
  e_pxc_cert_verify_web = 1,    // web of trust
  e_pxc_cert_verify_none = 2    // no verification
};

enum pxc_cert_verify_state {
  e_pxc_cert_invalid = -1,
  e_pxc_cert_unverified = 0,
  e_pxc_cert_verified = 1,
};

// helper structure for returning all of the digests needed for TOFU verification
// these are expected to be allocated via some malloc-like function (e.g. strdup)
struct pxc_x509_digests {
  char* cert_digest;
  char* pubkey_digest;
};

// pxc_ssl_aux_data contains data that will be stored in the SSL object for
// (mostly) certificate verification
struct pxc_ssl_aux_data {
  char const*                 hostname;
  char                        ip_port[INET6_ADDRSTRLEN + 1 + 5 + 1]; // enough for IPv6:port\0
  FILE*                       header_stream;

  // set in the verification
  enum pxc_cert_verify_mode   verify_mode;
  enum pxc_cert_verify_state  verify_state;
  struct pxc_x509_digests     cert_digests;
  char const*                 verify_path;  // path for SSL certificate verification data
                                            // this means different things depending on verify_mode
                                            // e_pxc_cert_verify_tofu: a path to the directory
                                            //  containing tofu verification files.  likely
                                            //  $HOME/.pxc/known_certs
                                            // e_pxc_cert_verify_web: a path to a file containing
                                            //  trusted CA certificates.
};


char* g_progname = NULL; // the name of our program, i.e. argv[0]
int   pxc_ssl_aux_data_index = -1; // index for auxiliary data stored in our SSL objects


#ifndef PXC_VERSION
#define PXC_VERSION unknown version
#endif

#define XSTR(x) # x
#define STR(x) XSTR(x)

static void version(FILE* out) {
  fprintf(out, "pxc %s\n", STR(PXC_VERSION));
}

static void usage(FILE* out, int full) {
  fprintf(out, "usage: %s [-c cert_file] [-C path] [-h] [-k key_file] "
                         "[-o out_path] [-q] [-r] [-V tofu|web|none] host ...\n", g_progname);
  if (full) {
    fputc('\n', stderr);
    version(out);

    fprintf(out, "\npxc is a command line Gemini client, designed as a filter program "
                 "in the spirit of traditional UNIX tools like 'grep' and 'sed'.  by default pxc prints "
                 "informational messages, errors, and the Gemini header to stderr and the response "
                 "body to stdout.\n\n");
    fprintf(out,
        "  -c cert_file    optional, the client certificate file to use.  this may also be specified\n"
        "                  by setting the PXC_CERT environment variable.  the command line switch will\n"
        "                  override the environment variable if both are specified.  if specified,\n"
        "                  this switch must be accompanied by a client key (see -k).\n"
        "  -C verify_path  path to server certificate verification data.\n"
        "                  if using the default 'tofu' verification (see -V) then this is a path\n"
        "                  to the known certificates directory (defaults to $HOME/.pxc/known_certs).\n"
        "                  if using 'web' verification then this is a path to the trusted CA\n"
        "                  certificate store (defaults to openssl defaults).\n"
        "  -h              display this help message.\n"
        "  -H              write the Gemini header to the same stream as the response body.\n"
        "  -k key_file     optional, the client private key file to use.  this may also be specified by\n"
        "                  setting the PXC_KEY environment variable.  the command line switch will override\n"
        "                  the environment variable if both are specified.  if set, this switch must be\n"
        "                  accompanied by a client certificate (see -c).\n"
        "  -o out_path     write the response to this file instead of stdout.\n"
        "  -q              quiet mode, no error or informational messages are displayed.\n"
        "  -r              reads the Gemini request line from stdin instead of using a command line argument.\n"
        "                  note that the host still needs to be specified on the command line so that\n"
        "                  a connection can be made.\n"
        "  -V option       server certificate verification mode.\n"
        "                  may be 'tofu' (default), 'web', or 'none'.\n"
        "                  tofu: trust on first use\n"
        "                    a host's certificate information is saved to the known certificate database\n"
        "                    the first time a connection is made.  subsequent connections to the host\n"
        "                    compares the certificate information sent by the host with that\n"
        "                    stored in the known certificate database.  if it the two do not match\n"
        "                    then the connection is terminated.\n"
        "                  web: web of trust model.\n"
        "                    similar to \"standard\" https verification.  trusted certificates\n"
        "                    are loaded from a trusted certificate store and all server certificates\n"
        "                    must be signed by one of those certificates, otherwise the connection.\n"
        "                    to the host is terminated.\n"
        "                  none: no certificate validation is done.\n");
  }
}

// calculate the digests needed for TOFU verification
static struct pxc_x509_digests hash_x509(X509* cert) {
  struct pxc_x509_digests digests = { 0 };
  unsigned char digest_buf[EVP_MAX_MD_SIZE];
  EVP_MD const* hashfunc = EVP_sha256();

  if (!cert || !hashfunc)
    return digests;

  // we do this to reduce boilerplate code
  // this is just pairs of digest functions to perform on the cert and
  // destinations for the resulting hex-encoded digest strings
  struct target_digest {
    int (*dgstfunc)(const X509 *, const EVP_MD *, unsigned char *, unsigned int *);
    char** out;
    char const* desc; // description of operation, for error reporting
  };

  // define what we want to do
  struct target_digest target_ops[] = {
    { .dgstfunc = X509_digest, .out = &digests.cert_digest, .desc = "certificate digest" },
    { .dgstfunc = X509_pubkey_digest, .out = &digests.pubkey_digest, .desc = "pubkey digest" }
  };

  char const* hexchars = "0123456789abcdef";

  for (unsigned opidx = 0, nops = sizeof(target_ops) / sizeof(target_ops[0]); opidx < nops; ++opidx) {
    // clear out our buffer
    unsigned digest_buf_sz = sizeof(digest_buf);
    memset(digest_buf, 0, digest_buf_sz);

    if (target_ops[opidx].dgstfunc(cert, hashfunc, digest_buf, &digest_buf_sz) != 1) {
      fprintf(stderr,
          "could not calculate %s: %s\n",
          target_ops[opidx].desc, ERR_error_string(ERR_get_error(), NULL));
      break;
    }

    unsigned hex_str_len = 2 * digest_buf_sz + 1;
    char* hexencoded = (char*)malloc(hex_str_len);
    if (!hexencoded) {
      fprintf(stderr, "no memory\n");
      break;
    }

    char* outitr = hexencoded;
    for (unsigned i = 0; i < digest_buf_sz; ++i, outitr += 2) {
      unsigned b = digest_buf[i];
      outitr[0] = hexchars[(b & 0xF0) >> 4];
      outitr[1] = hexchars[(b & 0x0F)];
    }
    *outitr = '\0';

    *target_ops[opidx].out = hexencoded;
  }

  // check for errors
  if (!digests.cert_digest || !digests.pubkey_digest) {
    free(digests.cert_digest);
    free(digests.pubkey_digest);
    memset(&digests, 0, sizeof(digests));
  }

  return digests;
}

// this compares certificate and hashes against those found in a given file
// 'f' contains a FILE stream to an opened TOFU data file
// 'filename' is provided for informational purposes in error messages
// 'cert_digest' and 'pubkey_digest' are the digests of the server's
// certificate and public keys, as they were sent to the client
static enum pxc_cert_verify_state
check_tofu_file(FILE* f, char const* filename, char const* cert_digest, char const* pubkey_digest)
{

  struct file_line {
    char* text;
    char const* to_compare;
    char const* desc;
  };

  struct file_line lines[] = {
    { NULL, cert_digest, "certificate digest" },
    { NULL, pubkey_digest, "pubkey digest" }
  };

  enum pxc_cert_verify_state ret = e_pxc_cert_invalid;

  // make sure the file has all of the required parts before we do any verification
  for (unsigned i = 0, n = sizeof(lines) / sizeof(lines[0]); i < n; ++i) {
    char* line = NULL;
    size_t allocsz = 0;
    ssize_t linesz = getline(&line, &allocsz, f);

    if (linesz < 0) {
      int e = errno;
      fprintf(stderr, "could not read %s from file: %s\n", lines[i].desc, strerror(e));
      goto DONE;
    }

    // strip the newline
    if (linesz >= 1 && line[linesz - 1] == '\n')
      line[linesz - 1] = '\0';

    lines[i].text = line;
  }

  unsigned n_lines = sizeof(lines) / sizeof(lines[0]);
  unsigned matched_idx = n_lines;
  for (unsigned i = 0; i < n_lines; ++i) {
    int cmp = strcmp(lines[i].text, lines[i].to_compare);
    if (cmp == 0) {
      matched_idx = i;
      if (i == 0) // matched on certificate digest, everything is good
        ret = e_pxc_cert_verified;
      break;
    }
    fprintf(stderr, "TOFU: mismatch on %s.\n"
                    "  expected: %s\n"
                    "  got:      %s\n", lines[i].desc, lines[i].text, lines[i].to_compare);
  }

  // there was a mismatch
  if (matched_idx > 0 && matched_idx < n_lines) {
    fprintf(stderr, "TOFU: note, the certificate value for this host should be updated:\n"
                    "  echo \"%s\" > %s\n"
                    "  echo \"%s\" >> %s\n",
                    cert_digest, filename,
                    pubkey_digest, filename);
  } else if (matched_idx != 0) {
    fprintf(stderr, "TOFU: certificate failed verification\n");
  }

DONE:
  for (unsigned i = 0; i < n_lines; ++i) {
    free(lines[i].text);
  }

  return ret;
}

// worker function for verifying a certificate using the TOFU database
static enum pxc_cert_verify_state verify_tofu(struct pxc_ssl_aux_data const* auxdata) {
  char const* hostname = auxdata->hostname;
  if (!hostname || hostname[0] == '\0') {
    fprintf(stderr, "empty hostname, cannot check TOFU store\n");
    return e_pxc_cert_invalid;
  }

  // double-check that hostname contains no slashes
  if (strchr(hostname, '/') != NULL) {
    fprintf(stderr, "hostname %s contains a slash, this should not happen.  "
                    "refusing to accept this request\n", hostname);
    return e_pxc_cert_invalid;
  }

  if (!auxdata->verify_path) {
    fprintf(stderr, "TOFU: no known certificate directory specified, cannot check certificate\n");
    return e_pxc_cert_invalid;
  }

  // TODO what we may want to do is sha256 the filenames to keep some metadata
  // a little bit more private, maybe even naming them something like
  // <salt>:<sha256> to prevent pre-computing hostname hashes.  this also might
  // be 100% overkill so we'll leave it as plaintext for now.
  char filenamebuf[PATH_MAX] = { 0 };
  int r = snprintf(filenamebuf, sizeof(filenamebuf), "%s/%s", auxdata->verify_path, hostname);
  if (r >= sizeof(filenamebuf)) {
    fprintf(stderr, "bug: overflow writing pathname for %s/%s, pxc needs to be fixed\n",
                    auxdata->verify_path, hostname);
    return e_pxc_cert_invalid;
  }

  enum pxc_cert_verify_state ret = e_pxc_cert_invalid;

  // TODO use the symbolic constants here
  int fd = open(filenamebuf, O_RDONLY | O_CLOEXEC);

  if (fd >= 0) {
    FILE* f = fdopen(fd, "r");
    if (!f) {
      fprintf(stderr, "could not fdopen for %s: %s\n", filenamebuf, strerror(errno));
      close(fd);
      return e_pxc_cert_invalid;
    }

    ret = check_tofu_file(f, filenamebuf,
                          auxdata->cert_digests.cert_digest,
                          auxdata->cert_digests.pubkey_digest);
    fclose(f);
  } else if (errno == ENOENT) { // file did not exist
    fd = open(filenamebuf, O_WRONLY | O_CREAT | O_EXCL | O_CLOEXEC, 0600);
    if (fd < 0) {
      fprintf(stderr, "could not open TOFU certificate file %s for writing: %s\n", filenamebuf, strerror(errno));
      return e_pxc_cert_invalid;
    }

    FILE* f = fdopen(fd, "w");
    if (!f) {
      fprintf(stderr, "could not fdopen for %s: %s\n", filenamebuf, strerror(errno));
      close(fd);
      if (unlink(filenamebuf) != 0)
        fprintf(stderr, "WARNING: could not unlink TOFU file %s: %s\n", filenamebuf, strerror(errno));
      return e_pxc_cert_invalid;
    }

    fprintf(stderr, "creating a new TOFU entry for %s\n", hostname);
    if (fprintf(f, "%s\n%s\n", auxdata->cert_digests.cert_digest, auxdata->cert_digests.pubkey_digest) < 0) {
      fprintf(stderr, "WARNING: could not write TOFU file %s: %s\n", filenamebuf, strerror(errno));

      if (unlink(filenamebuf) != 0)
        fprintf(stderr, "WARNING: could not unlink TOFU file %s: %s\n", filenamebuf, strerror(errno));
    } else {
      ret = e_pxc_cert_verified;
    }
    fclose(f);
  } else { // some other error, e.g. bad perms
    fprintf(stderr, "TOFU: could not read TOFU certificate file %s: %s\n", filenamebuf, strerror(errno));
  }

  return ret;
}

// function called by OpenSSL while verifying the certificate chain.
static int cert_verify_cb(int verify, X509_STORE_CTX *ctx) {

  char subject_buf[256] = { 0 };
  EVP_PKEY* pubkey = NULL;

  if (pxc_ssl_aux_data_index < 0) {
    fprintf(stderr, "bug: bad SSL setup, no host data index,\n");
    return 0;
  }

  SSL* ssl = X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
  if (!ssl) {
    fprintf(stderr, "bug: no SSL object when verifying certificate\n");
    return 0;
  }

  struct pxc_ssl_aux_data* data = SSL_get_ex_data(ssl, pxc_ssl_aux_data_index);
  if (!data) {
    fprintf(stderr, "bug: no host data attached to SSL object\n");
    return 0;
  }

  // no verification, accept all certs
  if (data->verify_mode == e_pxc_cert_verify_none) {
    if (data->verify_state == e_pxc_cert_unverified) {
      data->verify_state = e_pxc_cert_verified; // just means that we don't need to do any more printing
      fprintf(stderr, "NOTE: skipping certificate checks\n");
    }
    return 1;
  }

  if (data->verify_mode == e_pxc_cert_verify_web) { // web of trust mode, use OpenSSL's methods
    if (data->verify_state == e_pxc_cert_unverified) {
      data->verify_state = e_pxc_cert_verified; // just means that we don't need to do any more printing

      STACK_OF(X509)* cert_chain = X509_STORE_CTX_get0_chain(ctx);
      int cert_chain_sz = sk_X509_num(cert_chain);
      if (cert_chain_sz > 0) {
        fprintf(stderr, "WEB: certificate chain:\n");
        for (int i = 0; i < cert_chain_sz; ++i) {
          X509* cert = sk_X509_value(cert_chain, 0);
          if (!cert)
            continue;

          X509_NAME_oneline(X509_get_subject_name(cert), subject_buf, sizeof(subject_buf));
          fprintf(stderr, "  %d subject: %s\n", i, subject_buf);

          X509_NAME_oneline(X509_get_issuer_name(cert), subject_buf, sizeof(subject_buf));
          fprintf(stderr, "  %d issuer:  %s\n", i, subject_buf);
        }
      } else {
        fprintf(stderr, "WEB: no certificate chain!\n");
      }

      if (!verify) {
        int err = X509_STORE_CTX_get_error(ctx);
        fprintf(stderr, "certificate verification failed: %s\n", X509_verify_cert_error_string(err));
      }
    }
    return verify;
  }

  if (data->verify_state == e_pxc_cert_unverified) {

    STACK_OF(X509)* cert_chain = X509_STORE_CTX_get0_chain(ctx);

    int cert_chain_sz = sk_X509_num(cert_chain);
    if (cert_chain_sz == 0) {
      fprintf(stderr, "TOFU: no certificate chain to check!\n");
      return 0;
    }
    
    X509* cert = sk_X509_value(cert_chain, 0);

    if (!cert) {
      fprintf(stderr, "no certificate?\n");
      return e_pxc_cert_invalid;
    }

    X509_NAME_oneline(X509_get_subject_name(cert), subject_buf, sizeof(subject_buf));

    int err = X509_STORE_CTX_get_error(ctx);
    if (err == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT)
      fprintf(stderr, "NOTE: self-signed certificate\n");

    fprintf(stderr, "certificate info: %s\n", subject_buf);

    // grab the public key and then decrement its reference immediately
    // to make error handling easier.  we could use X509_get0_pubkey
    // to get the same result but libressl gets rid of some of those functions so
    // let's not rely on it
    pubkey = X509_get_pubkey(cert);
    if (!pubkey) {
      fprintf(stderr, "no pubkey in certificate\n");
      return 0; // this should not happen
    }
    EVP_PKEY_free(pubkey); // see above re why this is here

    struct pxc_x509_digests digests = hash_x509(cert);
    if (!digests.cert_digest || !digests.pubkey_digest) {
      // these should always be NULL here but make sure they get free'd anyway
      free(digests.cert_digest);
      free(digests.pubkey_digest);
      return 0;
    }

    if (data->cert_digests.cert_digest)
      free(data->cert_digests.cert_digest);
    if (data->cert_digests.pubkey_digest)
      free(data->cert_digests.pubkey_digest);

    data->cert_digests = digests;

    fprintf(stderr, "certificate digest: %s\n"
        "pubkey digest:      %s\n",
        data->cert_digests.cert_digest ? data->cert_digests.cert_digest : "",
        data->cert_digests.pubkey_digest ? data->cert_digests.pubkey_digest : "");

    data->verify_state = verify_tofu(data);
  }

  return data->verify_state == e_pxc_cert_verified;
}

// function that is called to retrieve a password for a PEM key or certificate
// file (e.g. client private key)
//
// the EVP_* methods used here were pulled from PEM_def_callback in pem_lib.c
// in the openssl source
static int pxc_pem_pw_cb(char *buf, int num, int rwflag, void *userdata) {

  // rwflag == 1 means encryption, which we shouldn't be doing
  if (rwflag) {
    fprintf(stderr, "pxc should not be encrypting anything...bug?\n");
    return -1;
  }

  if (!userdata) {
    fprintf(stderr, "pxc needs a password but it should know which file it's decrypting...bug?\n");
    return -1;
  }

  char* prompt = NULL;
  int r = asprintf(&prompt, "Enter passphrase for %s: ", (char const*)userdata);
  if (r < 0 || !prompt) {
    fprintf(stderr, "could not allocate prompt buffer: %s\n", strerror(errno));
    return -1;
  }

  r = EVP_read_pw_string_min(buf, 0, num, prompt, rwflag);
  free(prompt);
  if (r != 0) {
    ERR_raise(ERR_LIB_PEM, PEM_R_PROBLEMS_GETTING_PASSWORD);
    memset(buf, 0, (unsigned int)num);
    return -1;
  }
  return strlen(buf);
}

// set up the SSL_CTX object we will use to derive SSL objects for Gemini requests
static void setup_ssl_ctx(SSL_CTX* ctx, char const* cert_file, char const* pkey_file) {
  SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, cert_verify_cb);

  // cert and key must be specified together
  if (!cert_file != !pkey_file) {
    fprintf(stderr, "a certificate (-c) and key (-k) must be specified together\n");
    exit(-1);
  }

  if (cert_file) {
    fprintf(stderr, "Using client cert %s, key %s\n", cert_file, pkey_file);
    SSL_CTX_set_default_passwd_cb(ctx, pxc_pem_pw_cb);

    SSL_CTX_set_default_passwd_cb_userdata(ctx, (void*)cert_file);
    int r = SSL_CTX_use_certificate_chain_file(ctx, cert_file);
    if (r != 1) {
      unsigned long e = ERR_get_error();
      fprintf(stderr, "could not use cert chain file %s: %s\n", cert_file, ERR_error_string(e, NULL));
      exit(-1);
    }

    SSL_CTX_set_default_passwd_cb_userdata(ctx, (void*)pkey_file);
    r = SSL_CTX_use_PrivateKey_file(ctx, pkey_file, SSL_FILETYPE_PEM);
    if (r != 1) {
      unsigned long e = ERR_get_error();
      fprintf(stderr, "could not load private key %s: %s\n", pkey_file, ERR_error_string(e, NULL));
      exit(-1);
    }

    // reset the userdata
    SSL_CTX_set_default_passwd_cb_userdata(ctx, NULL);
  }

  if (cert_file || pkey_file) {
    int r = SSL_CTX_check_private_key(ctx);
    if (r != 1) {
      fprintf(stderr, "warning: private key does not match certificate chain: %s\n",
              ERR_error_string(ERR_get_error(), NULL));
    }
  }
}

#define STR_OR_NULL(x) ((x) ? (x) : "")

// helper function, converts socket information to a printable string for e.g. notifying the user
static int addr_to_str(struct sockaddr_storage const* sa, char* bufip, size_t bufip_sz) {
  char addrbuf[INET6_ADDRSTRLEN + 1] = { 0 };
  unsigned short port = -1;

  switch (sa->ss_family) {
    case AF_INET :
      {
        struct sockaddr_in const* sin = (struct sockaddr_in const*)sa;
        if (1) {
          char const* r = inet_ntop(AF_INET, &sin->sin_addr, addrbuf, sizeof(addrbuf));
          if (r != addrbuf) {
            fprintf(stderr, "overflow, buffers need adjustment\n");
            exit(-1);
          }
        }

        port = ntohs(sin->sin_port);
      }
      break;
    case AF_INET6 :
      {
        struct sockaddr_in6 const* sin6 = (struct sockaddr_in6 const*)sa;
        if (1) {
          char const* rc = inet_ntop(AF_INET6, &sin6->sin6_addr, addrbuf, sizeof(addrbuf));
          if (rc != bufip) {
            fprintf(stderr, "overflow, buffers need adjustment\n");
            exit(-1);
          }
        }

        port = ntohs(sin6->sin6_port);
      }
      break;
    case AF_UNIX :
      // fallthrough, don't support this yet
    default :
      fprintf(stderr, "unsupported connection medium\n");
      exit(-1);
  }

  memset(bufip, 0, bufip_sz);

  int r = 0;
  r = snprintf(bufip, bufip_sz, "%s:%hu", addrbuf, port);

  return r < (int)bufip_sz;
}

// send a Gemini request to a host
// sock - the socket to communicate on
// ctx - the SSL_CTX to use to set up the SSL object/connection
// reqbuf - buffer containing the request
// send_stdin - ignored for now, for future Titan support
// auxdata - auxiliary data, contains peer information and certificate hashes
void send_request(int sock, SSL_CTX* ctx, char const* reqbuf, int send_stdin,
                  struct pxc_ssl_aux_data* auxdata)
{
  SSL* ssl = SSL_new(ctx);
  char buf[1027] = { 0 };
  int need_send = strlen(reqbuf);

  {
    int req_println = strlen(reqbuf);
    char const* cr = strchr(reqbuf, '\r');
    if (cr) {
      req_println = cr - reqbuf;
    } else {
      cr = strchr(reqbuf, '\n');
      if (cr)
        req_println = cr - reqbuf;
    }
    fprintf(stderr, "requesting '%.*s'\n", req_println, reqbuf);
  }

  if (!ssl) {
    unsigned long e = ERR_get_error();
    fprintf(stderr, "could not set up SSL: %s\n", ERR_error_string(e, NULL));
    goto DONE;
  }

  if (SSL_set_fd(ssl, sock) != 1) {
    unsigned long e = ERR_get_error();
    fprintf(stderr, "could not set fds: %s\n", ERR_error_string(e, NULL));
    goto DONE;
  }

  if (auxdata->hostname) {
    if (SSL_set_tlsext_host_name(ssl, auxdata->hostname) != 1) {
      unsigned long e = ERR_get_error();
      fprintf(stderr, "could not set fds: %s\n", ERR_error_string(e, NULL));
      goto DONE;
    }
  }

  {
    struct sockaddr_storage sa = { 0 };
    socklen_t salen = sizeof(sa);
    if (getpeername(sock, (struct sockaddr*)&sa, &salen) < 0) {
      fprintf(stderr, "could not getpeername: %s\n", strerror(errno));
      goto DONE;
    }

    int r = addr_to_str(&sa, &auxdata->ip_port[0], sizeof(auxdata->ip_port));
    if (!r) {
      fprintf(stderr, "overflow, buffer sizes need adjustment\n");
      goto DONE;
    }
  }

  SSL_set_ex_data(ssl, pxc_ssl_aux_data_index, auxdata);

  if (SSL_connect(ssl) != 1) {
    unsigned long e = ERR_get_error();
    fprintf(stderr, "could not SSL_connect: %s\n", ERR_error_string(e, NULL));
    goto DONE;
  }

  // we shouldn't need this anymore
  SSL_set_ex_data(ssl, pxc_ssl_aux_data_index, NULL);

  while (need_send > 0) {
    int r = SSL_write(ssl, reqbuf, need_send);
    if (r <= 0) {
      unsigned long e = ERR_get_error();
      fprintf(stderr, "could not SSL_write: %s\n", ERR_error_string(e, NULL));
      goto DONE;
    }
    need_send -= r;
    reqbuf += r;
  }

  int read_header = 0;
  int buf_fill = 0;
  while (1) {
    int r = SSL_read(ssl, &buf[buf_fill], sizeof(buf) - 1 - buf_fill);
    if (r <= 0) {
      int e = SSL_get_error(ssl, r);
      if (e == SSL_ERROR_ZERO_RETURN) {
        // done
        break;
      } else {
        fprintf(stderr, "error on read: %s\n", ERR_error_string(e, NULL));
        break;
      }
    }

    buf_fill += r;

    // if we haven't gotten a header yet then find out if we've received it
    if (!read_header) {
      char* crlf = memmem(buf, buf_fill, "\r\n", 2);
      if (!crlf) {
        if (buf_fill == sizeof(buf) - 1) {
          fprintf(stderr, "no header found\n");
          break;
        }
        continue;
      }

      // header received, print it to stderr and then move the remaining data
      // forward in the buffer
      read_header = 1;
      fprintf(auxdata->header_stream ? auxdata->header_stream : stderr, "%.*s\n", (int)(crlf - buf), buf);
      buf_fill -= crlf + 2 - buf; // subtract out the header length
      assert(buf_fill >= 0); // check for underflow
      memmove(buf, crlf + 2, buf_fill); // move the rest of the data forward
    }

    if (buf_fill != 0) {
      int wr = fwrite(buf, buf_fill, 1, stdout);
      if (wr != 1) {
        fprintf(stderr, "error writing data: %s\n", strerror(errno));
        break;
      }
      buf_fill = 0;
    }
  }

  // send a close-notify response
  SSL_shutdown(ssl);
DONE:
  SSL_free(ssl);
  close(sock);
}

// establish a connection to a host using whatever means works
static int connect_to_host(SSL_CTX* ctx, struct px_url* url, int send_stdin, struct pxc_ssl_aux_data* auxdata)
{
  struct addrinfo* addr;
  struct addrinfo hint = { .ai_family = AF_UNSPEC, .ai_socktype = SOCK_STREAM };

  int r = getaddrinfo(url->host, url->port, &hint, &addr);
  if (r != 0) {
    fprintf(stderr, "could not get address info for %s:%s: %s\n",
                    STR_OR_NULL(url->host),
                    STR_OR_NULL(url->port),
                    gai_strerror(r));
    return -1;
  }

  int sock = -1;
  for (struct addrinfo* itr = addr; itr; itr = itr->ai_next) {
    int s = socket(itr->ai_family, itr->ai_socktype | SOCK_CLOEXEC, itr->ai_protocol);
    if (s < 0) {
      if (!itr->ai_next)
        fprintf(stderr, "could not create socket: %s\n", strerror(errno));
      continue;
    }

    if (connect(s, itr->ai_addr, itr->ai_addrlen) >= 0) {
      sock = s;
      break;
    }

    close(s);
  }

  freeaddrinfo(addr);
  return sock;
}

#define SCHEME "%s"
#define USERINFO "%s"
#define HOSTSEP "%s"
#define HOST "%s"
#define PORTSEP "%s"
#define PORT "%s"
#define PATH "%s"
#define QUERYSEP "%s"
#define QUERY "%s"
#define FRAGMENTSEP "%s"
#define FRAGMENT "%s"

int main(int argc, char* argv[]) {
  struct {
    char* cert_file;
    char* pkey_file;
    char* out_path;
    int   input_mode; // this enables titan
    int   read_request; // whether the request should be cleaned up
    int   quiet; // redirects stderr to /dev/null
    int   merge_header;
    char* verify_path;
    enum pxc_cert_verify_mode verify_mode;
  } options = { .verify_mode = e_pxc_cert_verify_tofu };

  g_progname = strdup(argv[0]);
  if (!g_progname) // no memory?
    exit(-1);

  int euid = geteuid();

  if (euid == 0) {
    char* use_root = getenv("PXC_AS_ROOT");
    if (!use_root || strcmp(use_root, "1") != 0) {
      fprintf(stderr, "You should not run this program as root.  "
                      "If you want to run it as root anyway, set the environment variable PXC_AS_ROOT=1\n");
      exit(-1);
    }
  }

  int c;
  // TODO add 'i:' to getopt for input mode via Titan (WIP)
  while ((c = getopt(argc, argv, ":c:C:k:o:rhHqV:v")) >= 0) {
    switch (c) {
      case 'c' :
        if (options.cert_file)
          free(options.cert_file);
        options.cert_file = strdup(optarg);
        break;
      case 'C' :
        if (options.verify_path)
          free(options.verify_path);
        options.verify_path = strdup(optarg);
        break;
      case 'h':
        usage(stdout, 1);
        exit(0);
      case 'H':
        options.merge_header = 1;
        break;
      case 'i' :
        //options.input_mode = 1; // enables titan
        fprintf(stderr, "-i not implemented yet, sorry.  retry without it.\n");
        exit(-1);
        break;
      case 'k' :
        if (options.pkey_file)
          free(options.pkey_file);
        options.pkey_file = strdup(optarg);
        break;
      case 'o' :
        if (options.out_path)
          free(options.out_path);
        options.out_path = strdup(optarg);
        break;
      case 'r' :
        options.read_request = 1;
        break;
      case 'q' :
        options.quiet = 1;
        break;
      case 'V' :
        if (strcmp(optarg, "tofu") == 0) {
          options.verify_mode = e_pxc_cert_verify_tofu;
        } else if (strcmp(optarg, "web") == 0) {
          options.verify_mode = e_pxc_cert_verify_web;
        } else if (strcmp(optarg, "none") == 0) {
          options.verify_mode = e_pxc_cert_verify_none;
        } else {
          fprintf(stderr, "unknown option for -V: %s.  may be one of 'tofu', 'web', or 'none'\n", optarg);
          usage(stderr, 0);
          exit(-1);
        }
        break;
      case 'v' :
        version(stdout);
        exit(0);
      case ':' :
        fprintf(stderr, "missing argument for -%c\n", optopt);
        usage(stderr, 0);
        exit(-1);
      case '?' :
        fprintf(stderr, "unknown option -%c\n", optopt);
      default:
        usage(stderr, 0);
        exit(-1);
    }
  }

  if (options.quiet) {
    FILE* f = freopen("/dev/null", "w", stderr);
    if (!f) {
      fprintf(stderr, "could not open /dev/null: %s\n", strerror(errno));
      exit(-1);
    }
  }

  if (!options.cert_file) {
    char* cert_env = getenv("PXC_CERT");
    if (cert_env) {
      options.cert_file = strdup(cert_env);
      fprintf(stderr, "NOTE: using certificate from $PXC_CERT=%s\n", options.cert_file);
    }
  }

  if (!options.pkey_file) {
    char* key_env = getenv("PXC_KEY");
    if (key_env) {
      options.pkey_file = strdup(key_env);
      fprintf(stderr, "NOTE: using key from $PXC_KEY=%s\n", options.pkey_file);
    }
  }


  // no certs dir provided via -C, use ~/.pxc/known_certs
  // we do this using getpwuid
  // TODO does using getpwuid provide any advantage over $HOME?  if a
  // malicious actor can get the user to modify $HOME then they are probably
  // already doing a lot of damage anyway
  if (options.verify_path == NULL) {
    if (options.verify_mode == e_pxc_cert_verify_tofu) {
      struct passwd* pw = getpwuid(euid);
      if (!pw) {
        fprintf(stderr, "could not get passwd info for uid=%d "
            "(needed to find the known certificate directory)\n", euid);
        exit(-1);
      }

      if (!pw->pw_dir) {
        fprintf(stderr, "could not determine the location of $HOME, please use -C or -I\n");
        exit(-1);
      }

      char* certs_dir = NULL;
      if (asprintf(&certs_dir, "%s/" DEFAULT_KNOWN_CERTS_DIR, pw->pw_dir) < 0) {
        fprintf(stderr, "no memory\n");
        exit(-1);
      }
      options.verify_path = certs_dir;
    }
  }

  // get a slot to put our auxiliary data in
  pxc_ssl_aux_data_index = SSL_get_ex_new_index(0, "auxdata index", NULL, NULL, NULL);
  if (pxc_ssl_aux_data_index < 0) {
    fprintf(stderr, "could not create a host data index for SSL data\n");
    exit(-1);
  }

  SSL_CTX* ctx = SSL_CTX_new(TLS_client_method());
  if (!ctx) {
    fprintf(stderr, "couldn't make SSL_CTX\n");
    exit(-1);
  }

  if (options.verify_mode == e_pxc_cert_verify_tofu) {
    if (options.verify_path == NULL) {
      fprintf(stderr, "verification data path could be found, please set it using '-C' or "
          "disable certificate verification with '-V none'\n");
      exit(-1);
    }

    struct stat s;
    if (stat(options.verify_path, &s) < 0) {
      int e = errno;
      fprintf(stderr, "could not stat known certificate directory %s: %s\n",
          options.verify_path, strerror(e));
      switch (e) {
        case ENOENT :
          fprintf(stderr, "please run: umask 0077; mkdir -p %s\n", options.verify_path);
          break;
        case EACCES :
          fprintf(stderr, "please correct the permissions of the path\n");
      }
      exit(-1);
    }

    if ((s.st_mode & S_IFMT) != S_IFDIR) {
      fprintf(stderr, "TOFU known certificate directory %s is not a directory\n", options.verify_path);
      exit(-1);
    }

    // reduce the amount of work we need to do, everything happens the first
    // time our certificate verify callback happens
    SSL_CTX_set_verify_depth(ctx, 0);
  } else if (options.verify_mode == e_pxc_cert_verify_web) {
    int r = options.verify_path == NULL
            ? SSL_CTX_set_default_verify_store(ctx)
            : SSL_CTX_use_certificate_chain_file(ctx, options.verify_path);
    if (r != 1) {
      fprintf(stderr, "WEB: could not load trusted certificates: %s", ERR_error_string(ERR_get_error(), NULL));
      exit(-1);
    }
  }

  setup_ssl_ctx(ctx, options.cert_file, options.pkey_file);

  // make connections to each remaining host on the command line
  for (int i = optind; i < argc; ++i) {
    char const* reqstr = argv[i];
    struct px_url url = { 0 };
    if (px_url_parse(&url, reqstr, strlen(reqstr)) != 0) {
      fprintf(stderr, "couldn't parse %s\n", argv[i]);
      continue;
    }

    char reqbuf[1027] = { 0 };

    if (!url.scheme)
      url.scheme = strdup("gemini://");

    if (!options.read_request) { // take the gemini request from the command line argument
      // build the request url.  basically it's an snprintf with a bunch of '%s'
      // this construct is stupid and bad, I'm sorry.
      int r = snprintf(reqbuf, sizeof(reqbuf) - 3 /* 1024 */,
          SCHEME USERINFO HOSTSEP HOST PORTSEP PORT PATH QUERYSEP QUERY FRAGMENTSEP FRAGMENT,
          STR_OR_NULL(url.scheme) /* SCHEME*/,
          STR_OR_NULL(url.userinfo) /* USERINFO */,
          url.userinfo ? "@" : "" /* HOSTSEP */,
          STR_OR_NULL(url.host) /* HOST */,
          url.port ? ":" : "" /* PORTSEP */,
          STR_OR_NULL(url.port) /* PORT */,
          STR_OR_NULL(url.path) /* PATH */,
          url.query ? "?" : "" /* QUERYSEP */,
          STR_OR_NULL(url.query) /* QUERY */,
          url.fragment ? "#" : "" /* FRAGMENTSEP */,
          STR_OR_NULL(url.fragment) /* FRAGMENT */);
      if (r >= sizeof(reqbuf) - 3) {
        fprintf(stderr, "cleaned request for %s too long (%d bytes)\n", reqstr, r);
        continue;
      }

      reqbuf[r] = '\r';
      reqbuf[r+1] = '\n';
      reqbuf[r+2] = '\0';
    } else { // take the gemini request from stdin.  useful if you want to send non-standard formats
      char* readreq = NULL;
      size_t alloc_sz = 0;
      ssize_t line_sz = getline(&readreq, &alloc_sz, stdin);
      if (line_sz < 0) {
        int e = errno;
        free(readreq);
        fprintf(stderr, "could not read request from stdin: %s\n", strerror(e));
        continue;
      }

      if (line_sz >= sizeof(reqbuf)) {
        free(readreq);
        fprintf(stderr, "request from stdin is %ld bytes long, must be 1024 or fewer.\n", line_sz);
        continue;
      }
      strncpy(reqbuf, readreq, sizeof(reqbuf) - 1);
      reqbuf[line_sz] = '\0';
      free(readreq);
    }

    // default to port 1965, standard Gemini port, if one was not specified
    if (!url.port)
      url.port = strdup("1965"); // default

    // if an output file was specified, open it up as stdout
    if (options.out_path) {
      FILE* f = freopen(options.out_path, "w", stdout);
      if (!f) {
        int e = errno;
        fprintf(stderr, "could not open %s for writing: %s\n", options.out_path, strerror(e));
        exit(-1);
      }
    }

    struct pxc_ssl_aux_data auxdata = {
      .hostname = url.host,
      .verify_mode = options.verify_mode,
      .verify_state = e_pxc_cert_unverified,
      .verify_path = options.verify_path };

    if (options.merge_header)
      auxdata.header_stream = stdout;
    else
      auxdata.header_stream = stderr;

    int sock = connect_to_host(ctx, &url, options.input_mode, &auxdata);
    if (sock >= 0) {
      send_request(sock, ctx, reqbuf, options.input_mode, &auxdata);
    } else {
      fprintf(stderr, "could not connect to host %s\n", url.host);
    }
    px_url_reset(&url);
    free(auxdata.cert_digests.cert_digest);
    free(auxdata.cert_digests.pubkey_digest);
  }
  SSL_CTX_free(ctx);

  char** to_free[] = {
    &options.cert_file,
    &options.pkey_file,
    &options.out_path,
    &options.verify_path,
    &g_progname
  };

  for (unsigned i = 0, ni = sizeof(to_free) / sizeof(to_free[0]); i < ni; ++i) {
    free(*to_free[i]);
    *to_free[i] = NULL;
  }
}
